@@ -2,6 +2,8 @@
 Contains the primary optimization and contraction routines.
 """
 
+from collections import namedtuple
+
 import numpy as np
 
 from . import backends
@@ -15,9 +17,85 @@ from . import sharing
 __all__ = ["contract_path", "contract", "format_const_einsum_str", "ContractExpression", "shape_only", "shape_only"]
 
 
+class PathInfo(object):
+    """A printable object to contain information about a contraction path.
+
+    Attributes
+    ----------
+    naive_cost : int
+        The estimate FLOP cost of a naive einsum contraction.
+    opt_cost : int
+        The estimate FLOP cost of this optimized contraction path.
+    largest_intermediate : int
+        The number of elements in the largest intermediate array that will be
+        produced during the contraction.
+    """
+
+    def __init__(self, contraction_list, input_subscripts, output_subscript,
+                 indices, scale_list, naive_cost, opt_cost, size_list):
+        self.contraction_list = contraction_list
+        self.input_subscripts = input_subscripts
+        self.output_subscript = output_subscript
+        self.indices = indices
+        self.scale_list = scale_list
+        self.naive_cost = naive_cost
+        self.opt_cost = opt_cost
+        self.largest_intermediate = max(size_list)
+
+    def __repr__(self):
+        from decimal import Decimal
+
+        # naive costs / speedups can easily reach >~ 1e308, need exact arithmetic
+        naive_cost = Decimal(self.naive_cost)
+        opt_cost = Decimal(self.opt_cost)
+        speedup = naive_cost / opt_cost
+        largest_intermediate = Decimal(self.largest_intermediate)
+
+        # Return the path along with a nice string representation
+        overall_contraction = self.input_subscripts + "->" + self.output_subscript
+        header = ("scaling", "BLAS", "current", "remaining")
+
+        path_print = [
+            u"  Complete contraction:  {}\n".format(overall_contraction),
+            u"         Naive scaling:  {}\n".format(len(self.indices)),
+            u"     Optimized scaling:  {}\n".format(max(self.scale_list)),
+            u"      Naive FLOP count:  {:.3e}\n".format(naive_cost),
+            u"  Optimized FLOP count:  {:.3e}\n".format(opt_cost),
+            u"   Theoretical speedup:  {:3.3f}\n".format(speedup),
+            u"  Largest intermediate:  {:.3e} elements\n".format(largest_intermediate),
+            u"-" * 80 + "\n",
+            u"{:>6} {:>11} {:>22} {:>37}\n".format(*header),
+            u"-" * 80
+        ]
+
+        for n, contraction in enumerate(self.contraction_list):
+            inds, idx_rm, einsum_str, remaining, do_blas = contraction
+            remaining_str = ",".join(remaining) + "->" + self.output_subscript
+            path_run = (self.scale_list[n], do_blas, einsum_str, remaining_str)
+            path_print.append(u"\n{:>4} {:>14} {:>22} {:>37}".format(*path_run))
+
+        return "".join(path_print)
+
+
+def _choose_memory_arg(memory_limit, size_list):
+    if memory_limit == 'max_input':
+        return max(size_list)
+
+    if memory_limit is None:
+        return None
+
+    if memory_limit < 1:
+        if memory_limit == -1:
+            return None
+        else:
+            raise ValueError("Memory limit must be larger than 0, or -1")
+
+    return int(memory_limit)
+
+
 def contract_path(*operands, **kwargs):
     """
-    Evaluates the lowest cost einsum-like contraction order.
+    Find a contraction order 'path', without performing the contraction.
 
     Parameters
     ----------
@@ -25,30 +103,34 @@ def contract_path(*operands, **kwargs):
         Specifies the subscripts for summation.
     *operands : list of array_like
         These are the arrays for the operation.
-    path : bool or list, optional (default: ``auto``)
+    optimize : str, list or bool, optional (default: ``auto``)
         Choose the type of path.
 
         - if a list is given uses this as the path.
-        - 'greedy' An algorithm that chooses the best pair contraction
-          at each step. Scales cubically with the number of terms in the
-          contraction.
-        - 'optimal' An algorithm that tries all possible ways of
-          contracting the listed tensors. Scales exponentially with
-          the number of terms in the contraction.
-        - 'auto' Use the optimal approach for a small number of terms (currently
-          4 or less), else use the greedy approach.
+        - ``'optimal'`` An algorithm that explores all possible ways of
+          contracting the listed tensors. Scales factorially with the number of
+          terms in the contraction.
+        - ``'branch-all'`` An algorithm like optimal but that restricts itself
+          to searching 'likely' paths. Still scales factorially.
+        - ``'branch-2'`` An even more restricted version of 'branch-all' that
+          only searches the best two options at each step. Scales exponentially
+          with the number of terms in the contraction.
+        - ``'greedy'`` An algorithm that heuristically chooses the best pair
+          contraction at each step.
+        - ``'auto'`` Choose the best of the above algorithms whilst aiming to
+          keep the path finding time below 1ms.
 
     use_blas : bool
         Use BLAS functions or not
-    memory_limit : int, optional (default: largest input or output array size)
+    memory_limit : int, optional (default: None)
         Maximum number of elements allowed in intermediate arrays.
 
     Returns
     -------
     path : list of tuples
         The einsum path
-    string_repr : str
-        A printable representation of the path
+    PathInfo : str
+        A printable object containing various information about the path found.
 
     Notes
     -----
@@ -111,12 +193,18 @@ def contract_path(*operands, **kwargs):
     """
 
     # Make sure all keywords are valid
-    valid_contract_kwargs = ['path', 'memory_limit', 'einsum_call', 'use_blas']
+    valid_contract_kwargs = ['optimize', 'path', 'memory_limit', 'einsum_call', 'use_blas']
     unknown_kwargs = [k for (k, v) in kwargs.items() if k not in valid_contract_kwargs]
     if len(unknown_kwargs):
-        raise TypeError("einsum_path: Did not understand the following kwargs: %s" % unknown_kwargs)
+        raise TypeError("einsum_path: Did not understand the following kwargs: {}".format(unknown_kwargs))
+
+    if 'path' in kwargs:
+        import warnings
+        warnings.warn("The 'path' keyword argument is deprecated in favor of 'optimize'.", DeprecationWarning)
+        path_type = kwargs.pop('path')
+    else:
+        path_type = kwargs.pop('optimize', 'auto')
 
-    path_type = kwargs.pop('path', 'auto')
     memory_limit = kwargs.pop('memory_limit', None)
 
     # Hidden option, only einsum should call this
@@ -139,8 +227,8 @@ def contract_path(*operands, **kwargs):
         sh = input_shps[tnum]
 
         if len(sh) != len(term):
-            raise ValueError("Einstein sum subscript %s does not contain the "
-                             "correct number of indices for operand %d." % (input_subscripts[tnum], tnum))
+            raise ValueError("Einstein sum subscript '{}' does not contain the "
+                             "correct number of indices for operand {}.".format(input_subscripts[tnum], tnum))
         for cnum, char in enumerate(term):
             dim = sh[cnum]
 
@@ -149,25 +237,14 @@ def contract_path(*operands, **kwargs):
                 if dimension_dict[char] == 1:
                     dimension_dict[char] = dim
                 elif dim not in (1, dimension_dict[char]):
-                    raise ValueError("Size of label '%s' for operand %d (%d) "
-                                     "does not match previous terms (%d)." % (char, tnum, dimension_dict[char], dim))
+                    raise ValueError("Size of label '{}' for operand {} ({}) does not match previous "
+                                     "terms ({}).".format(char, tnum, dimension_dict[char], dim))
             else:
                 dimension_dict[char] = dim
 
     # Compute size of each input array plus the output array
     size_list = [helpers.compute_size_by_dict(term, dimension_dict) for term in input_list + [output_subscript]]
-    out_size = max(size_list)
-
-    if memory_limit is None:
-        memory_arg = out_size
-    else:
-        if memory_limit < 1:
-            if memory_limit == -1:
-                memory_arg = int(1e20)
-            else:
-                raise ValueError("Memory limit must be larger than 0, or -1")
-        else:
-            memory_arg = int(memory_limit)
+    memory_arg = _choose_memory_arg(memory_limit, size_list)
 
     num_ops = len(input_list)
 
@@ -187,15 +264,18 @@ def contract_path(*operands, **kwargs):
     elif num_ops == 2:
         # Nothing to be optimized
         path = [(0, 1)]
-    elif indices == output_set:
-        # If no rank reduction leave it to einsum
-        path = [tuple(range(num_ops))]
     elif path_type == "optimal" or (path_type == "auto" and num_ops <= 4):
         path = paths.optimal(input_sets, output_set, dimension_dict, memory_arg)
-    elif path_type in ("greedy", "opportunistic", "auto"):
+    elif path_type == 'branch-all' or (path_type == "auto" and num_ops <= 6):
+        path = paths.branch(input_sets, output_set, dimension_dict, memory_arg, nbranch=None)
+    elif path_type == 'branch-2' or (path_type == "auto" and num_ops <= 8):
+        path = paths.branch(input_sets, output_set, dimension_dict, memory_arg, nbranch=2)
+    elif path_type == 'branch-1' or (path_type == "auto" and num_ops <= 14):
+        path = paths.branch(input_sets, output_set, dimension_dict, memory_arg, nbranch=1)
+    elif path_type in ("auto", "greedy", "eager", "opportunistic"):
         path = paths.greedy(input_sets, output_set, dimension_dict, memory_arg)
     else:
-        raise KeyError("Path name %s not found" % path_type)
+        raise KeyError("Path name '{}' not found".format(path_type))
 
     cost_list = []
     scale_list = []
@@ -247,26 +327,8 @@ def contract_path(*operands, **kwargs):
     if einsum_call_arg:
         return operands, contraction_list
 
-    # Return the path along with a nice string representation
-    overall_contraction = input_subscripts + "->" + output_subscript
-    header = ("scaling", "BLAS", "current", "remaining")
-
-    path_print = "  Complete contraction:  %s\n" % overall_contraction
-    path_print += "         Naive scaling:  %d\n" % len(indices)
-    path_print += "     Optimized scaling:  %d\n" % max(scale_list)
-    path_print += "      Naive FLOP count:  %.3e\n" % naive_cost
-    path_print += "  Optimized FLOP count:  %.3e\n" % opt_cost
-    path_print += "   Theoretical speedup:  %3.3f\n" % (naive_cost / float(opt_cost))
-    path_print += "  Largest intermediate:  %.3e elements\n" % max(size_list)
-    path_print += "-" * 80 + "\n"
-    path_print += "%6s %11s %22s %37s\n" % header
-    path_print += "-" * 80
-
-    for n, contraction in enumerate(contraction_list):
-        inds, idx_rm, einsum_str, remaining, do_blas = contraction
-        remaining_str = ",".join(remaining) + "->" + output_subscript
-        path_run = (scale_list[n], do_blas, einsum_str, remaining_str)
-        path_print += "\n%4d %14s %22s %37s" % path_run
+    path_print = PathInfo(contraction_list, input_subscripts, output_subscript,
+                          indices, scale_list, naive_cost, opt_cost, size_list)
 
     return path, path_print
 
@@ -294,16 +356,17 @@ def _einsum(*operands, **kwargs):
     return fn(einsum_str, *operands, **kwargs)
 
 
+def _default_transpose(x, axes):
+    #  most libraries implement a method version
+    return x.transpose(axes)
+
+
 @sharing.transpose_cache_wrap
 def _transpose(x, axes, backend='numpy'):
     """Base transpose.
     """
-    try:
-        return x.transpose(axes)
-    except (AttributeError, TypeError):
-        # some libraries don't implement method version
-        fn = backends.get_func('transpose', backend)
-        return fn(x, axes)
+    fn = backends.get_func('transpose', backend, _default_transpose)
+    return fn(x, axes)
 
 
 @sharing.tensordot_cache_wrap
@@ -339,22 +402,32 @@ def contract(*operands, **kwargs):
         The casting procedure for operations of different dtype, see np.einsum.
     use_blas : bool
         Do you use BLAS for valid operations, may use extra memory for more intermediates.
-    optimize : bool, str, or list, optional (default: ``greedy``)
+    optimize : str, list or bool, optional (default: ``auto``)
         Choose the type of path.
 
         - if a list is given uses this as the path.
-        - 'greedy' An algorithm that chooses the best pair contraction
-          at each step. Scales cubically with the number of terms in the
-          contraction.
-        - 'optimal' An algorithm that tries all possible ways of
-          contracting the listed tensors. Scales exponentially with
-          the number of terms in the contraction.
+        - ``'optimal'`` An algorithm that explores all possible ways of
+          contracting the listed tensors. Scales factorially with the number of
+          terms in the contraction.
+        - ``'branch-all'`` An algorithm like optimal but that restricts itself
+          to searching 'likely' paths. Still scales factorially.
+        - ``'branch-2'`` An even more restricted version of 'branch-all' that
+          only searches the best two options at each step. Scales exponentially
+          with the number of terms in the contraction.
+        - ``'greedy'`` An algorithm that heuristically chooses the best pair
+          contraction at each step.
+        - ``'auto'`` Choose the best of the above algorithms whilst aiming to
+          keep the path finding time below 1ms.
 
-    memory_limit : int or None (default : None)
-        The upper limit of the size of tensor created, by default, this will be
+    memory_limit : {None, int, 'max_input'} (default: None)
         Give the upper bound of the largest intermediate tensor contract will build.
-        By default (None) will size the ``memory_limit`` as the largest input tensor.
-        Users can also specify ``-1`` to allow arbitrarily large tensors to be built.
+
+        - None or -1 means there is no limit
+        - 'max_input' means the limit is set as largest input tensor
+        - a positive integer is taken as an explicit limit on the number of elements
+
+        The default is None. Note that imposing a limit can make contractions
+        exponentially slower to perform.
     backend : str, optional (default: ``numpy``)
         Which library to use to perform the required ``tensordot``, ``transpose``
         and ``einsum`` calls. Should match the types of arrays supplied, See
@@ -388,7 +461,7 @@ def contract(*operands, **kwargs):
     """
     optimize_arg = kwargs.pop('optimize', True)
     if optimize_arg is True:
-        optimize_arg = 'greedy'
+        optimize_arg = 'auto'
 
     valid_einsum_kwargs = ['out', 'dtype', 'order', 'casting']
     einsum_kwargs = {k: v for (k, v) in kwargs.items() if k in valid_einsum_kwargs}
@@ -407,14 +480,14 @@ def contract(*operands, **kwargs):
     # Make sure remaining keywords are valid for einsum
     unknown_kwargs = [k for (k, v) in kwargs.items() if k not in valid_einsum_kwargs]
     if len(unknown_kwargs):
-        raise TypeError("Did not understand the following kwargs: %s" % unknown_kwargs)
+        raise TypeError("Did not understand the following kwargs: {}".format(unknown_kwargs))
 
     if gen_expression:
         full_str = operands[0]
 
     # Build the contraction list and operand
     operands, contraction_list = contract_path(
-        *operands, path=optimize_arg, memory_limit=memory_limit, einsum_call=True, use_blas=use_blas)
+        *operands, optimize=optimize_arg, memory_limit=memory_limit, einsum_call=True, use_blas=use_blas)
 
     # check if performing contraction or just building expression
     if gen_expression:
@@ -630,13 +703,13 @@ class ContractExpression:
 
         if kwargs:
             raise ValueError("The only valid keyword arguments to a `ContractExpression` "
-                             "call are `out=` or `backend=`. Got: %s." % kwargs)
+                             "call are `out=` or `backend=`. Got: {}.".format(kwargs))
 
         correct_num_args = self._full_num_args if evaluate_constants else self.num_args
 
         if len(arrays) != correct_num_args:
-            raise ValueError("This `ContractExpression` takes exactly %s array arguments "
-                             "but received %s." % (self.num_args, len(arrays)))
+            raise ValueError("This `ContractExpression` takes exactly {} array arguments "
+                             "but received {}.".format(self.num_args, len(arrays)))
 
         if self._constants_dict and not evaluate_constants:
             # fill in the missing non-constant terms with newly supplied arrays
@@ -657,7 +730,7 @@ class ContractExpression:
             original_msg = str(err.args) if err.args else ""
             msg = ("Internal error while evaluating `ContractExpression`. Note that few checks are performed"
                    " - the number and rank of the array arguments must match the original expression. "
-                   "The internal error was: '%s'" % original_msg, )
+                   "The internal error was: '{}'".format(original_msg), )
             err.args = msg
             raise
 
@@ -669,20 +742,23 @@ class ContractExpression:
         return "<ContractExpression('{}'{})>".format(self.contraction, constants_repr)
 
     def __str__(self):
-        s = self.__repr__()
+        s = [self.__repr__()]
         for i, c in enumerate(self.contraction_list):
-            s += "\n  %i.  " % (i + 1)
-            s += "'%s'" % c[2] + (" [%s]" % c[-1] if c[-1] else "")
+            s.append("\n  {}.  ".format(i + 1))
+            s.append("'{}'".format(c[2]) + (" [{}]".format(c[-1]) if c[-1] else ""))
         if self.einsum_kwargs:
-            s += "\neinsum_kwargs=%s" % self.einsum_kwargs
-        return s
+            s.append("\neinsum_kwargs={}".format(self.einsum_kwargs))
+        return "".join(s)
+
+
+Shaped = namedtuple('Shaped', ['shape'])
 
 
 def shape_only(shape):
     """Dummy ``numpy.ndarray`` which has a shape only - for generating
     contract expressions.
     """
-    return np.broadcast_to(np.nan, shape)
+    return Shaped(shape)
 
 
 def contract_expression(subscripts, *shapes, **kwargs):
@@ -751,8 +827,11 @@ def contract_expression(subscripts, *shapes, **kwargs):
 
     for arg in ('out', 'backend'):
         if kwargs.get(arg, None) is not None:
-            raise ValueError("'%s' should only be specified when calling a "
-                             "`ContractExpression`, not when building it." % arg)
+            raise ValueError("'{}' should only be specified when calling a "
+                             "`ContractExpression`, not when building it.".format(arg))
+
+    if not isinstance(subscripts, compat.strings):
+        subscripts, shapes = parser.convert_interleaved_input((subscripts,) + shapes)
 
     kwargs['_gen_expression'] = True
 
